/* written by John Doug Reynolds, May 2002 */

#include "rgsfilter.h"
#include <ExposureMapper.h>
#include <EventAttributes.h>
#include <DataFilter.h>
#include <OGIPgti.h>
#include <cstdio>
#include <param.h>

namespace {
   using namespace rgsPixelsLib;
   using namespace std;
   using std::string; using std::sprintf;

   class ExposureCombiner {
   public:
      ~ExposureCombiner();

      ExposureCombiner( const string &filename, const DataSet *eventset,
			const GridSpan &beta, const GridSpan &xdsp );

      void load( const PixelMap<float>&,
		 const char *extname, unsigned ccd, unsigned node );

   private:
      struct DataSet* _dataset;
      struct Array* _array;
      struct ArrayData* _arrayData;
      struct PixelMap<float> _map;
      struct Span _column, _nodeid;
   };
}

static string newname( string model,
		       const string &ptype, const string &alternate );

static void addExpMaps( struct DataSet*,
			float driftbinsize, const string &combmap, const string spectrumbinning );

void rgsfilter()
{

	// We are going to check if RGS2 is in sigle readout mode. Therefore, we have to
	// open the dataset and check if SINGLERO is true (1) or false (0)
	string dsName = stringParameter("mergedset");
	string fsName = stringParameter("evlist");
	if ( fsName == "" ) fsName = newname(dsName,"EVENLI","evlist");
	struct DataSet *dataset = dataSetServer->clone(dsName,fsName,Dal::Modify);


	int rejflags;
   {
      using namespace EventAttributes;

      rejflags = CLEAR_ALL_FLAGS;
      std::vector<string>::const_iterator itr;
      std::vector<string> evattrList = stringListParameter("rejflags");

      for ( itr = evattrList.begin(); itr != evattrList.end(); ++itr )
	 if ( *itr == "BAD_SHAPE" )
	    setEventAttribute( rejflags, BAD_SHAPE );
	 else if ( *itr == "ON_NODE_INTERFACE" )
	 {
		 //Check if we are dealing with RGS2 single readout mode.
		 //If this is the case, we automatically disable this flag.
		 if ( (dataset->hasAttribute("SINGLERO")) &&
				 (dataset->attribute("SINGLERO")->asInt() == 1) )
		 {
			 errstr << "Single readout mode enabled. " <<
					 "ON_NODE_INTERFACE flag will be automatically disabled. " <<
					 Sas::message(Msg::App,Msg::Verbose) ;
		 }
		 else
			 setEventAttribute( rejflags, ON_NODE_INTERFACE );
	 }
	 else if ( *itr == "ON_BADPIX" )
	    setEventAttribute( rejflags, ON_BADPIX );
	 else if ( *itr == "NEXT_TO_BADPIX" )
	    setEventAttribute( rejflags, NEXT_TO_BADPIX );
	 else if ( *itr == "ON_WINDOW_BORDER" )
	    setEventAttribute( rejflags, ON_WINDOW_BORDER );
	 else if ( *itr == "BELOW_ACCEPTANCE" )
	    setEventAttribute( rejflags, BELOW_ACCEPTANCE );
	 else
	    Sas::error("badParameter");
   }


   // complete the STDGTI tables

   struct GtiList auxGTI(true); // all time is good

   std::vector<string>::const_iterator itr;
   std::vector<string> auxgtitables = stringListParameter("auxgtitables");
   for ( itr = auxgtitables.begin(); itr != auxgtitables.end(); ++itr ) {
      struct InOGIPgtiTable auxTable(*itr);
      struct GtiList gti;
      auxTable >> gti;
      auxGTI &= gti;
   }

   unsigned ccd;
   for ( ccd = 1; ccd <= 9; ++ccd ) {
      char extname[9];
      sprintf( extname, "STDGTI0%u", ccd );
      if ( !dataset->hasBlock(extname) ) continue;
      struct Table *gtab = dataset->table(extname);

      // align auxGTI with valid frames, then merge with STDGTI

      sprintf( extname, "EXPOSU0%u", ccd );
      struct Table *xtab = dataset->table(extname);
      DalVecD time( xtab->column("TIME") );
      DalVecE timedel( xtab->column("TIMEDEL") );
      DalVecE aspcdsp( xtab->column("ASPCDSP") );

      struct GtiList stdGTI, alignedGTI;
      GtiList::const_iterator itr = auxGTI.begin();
      for ( unsigned i = 0; i < xtab->rows(); ++i )
	 if ( !aspcdsp.isNull(i) ) {
	    double t = time[i], d = timedel[i] / 2;
	    struct Gti gti(t-d,t+d);
	    while ( itr != auxGTI.end() && itr->stop() < gti.stop() ) ++itr;
	    if ( itr == auxGTI.end() ) break;
	    if ( itr->start() <= gti.start() ) alignedGTI.push_back(gti);
	 }

      struct InOGIPgtiTable stdgtiIn(gtab);
      struct OutOGIPgtiTable stdgtiOut(gtab);
      stdgtiIn >> stdGTI;
      stdgtiOut << (stdGTI &= alignedGTI);
   }

   // filter the EVENTS table

   struct Table *etab = dataset->table("EVENTS");

   if ( !etab->hasAttribute("REJFLAGS") )
      *etab->addAttribute("REJFLAGS","flags already filtered","XMMEA") = 0;

   struct Attribute &rejflagsAttr = *etab->attribute("REJFLAGS");
   rejflagsAttr = rejflagsAttr.asInt() | rejflags;

   std::list<string> gtis;
   for ( ccd = 1; ccd <= 9; ++ccd ) {
      char extname[9], subexpr[64];
      sprintf( extname, "STDGTI0%u", ccd );
      if ( !dataset->hasBlock(extname) ) continue;
      sprintf( subexpr, "(CCDNR == %u && GTI([%s],TIME))", ccd, extname );
      gtis.push_back( subexpr );
   }

   string expression = "(FLAG & #REJFLAGS == 0)";
   if ( gtis.size() ) {
      std::list<string>::const_iterator itr = gtis.begin();
      expression += " && (" + *itr++;
      while ( itr != gtis.end() ) expression += " || " + *itr++;
      expression += ")";
   }

   DataFilter(expression).doFilter(*etab);

   // filter the EXPOSURE and REJPIX tables

   for ( ccd = 1; ccd <= 9; ++ccd ) {
      char xtabName[9], rtab0Name[9], rtab1Name[9];

      sprintf( xtabName, "EXPOSU0%u", ccd );
      if ( !dataset->hasBlock(xtabName) ) continue;

      char expression[32];
      struct Table *xtab = dataset->table(xtabName);
      sprintf( expression, "GTI([STDGTI0%u],TIME)", ccd );
      DataFilter(expression).doFilter(*xtab);

      sprintf( rtab0Name, "REJPIX0%u", ccd );
      sprintf( rtab1Name, "REJPIX1%u", ccd );

      unsigned r0Count = 0, r1Count = 0;
      struct Table *rtab0 = dataset->table(rtab0Name);
      struct Table *rtab1 = dataset->table(rtab1Name);
      {
	 const DalVecJ xFrame( xtab->column("FRAME") );
	 const DalVecJ r0Frame( rtab0->column("FRAME") );
	 const DalVecJ r1Frame( rtab1->column("FRAME") );

	 unsigned xItr = 0, r0Itr = 0, r1Itr = 0;
	 while ( xItr < xtab->rows() ) {
	    int frame = xFrame[xItr++];

	    while ( r0Itr < rtab0->rows() && r0Frame[r0Itr] < frame ) ++r0Itr;
	    while ( r0Itr < rtab0->rows() && r0Frame[r0Itr] == frame ) {
	       if ( r0Itr != r0Count ) rtab0->copyRows(r0Itr,r0Count);
	       ++r0Itr, ++r0Count;
	    }

	    while ( r1Itr < rtab1->rows() && r1Frame[r1Itr] < frame ) ++r1Itr;
	    while ( r1Itr < rtab1->rows() && r1Frame[r1Itr] == frame ) {
	       if ( r1Itr != r1Count ) rtab1->copyRows(r1Itr,r1Count);
	       ++r1Itr, ++r1Count;
	    }
	 }
      }
      rtab0->deleteRows( r0Count, rtab0->rows() - r0Count );
      rtab1->deleteRows( r1Count, rtab1->rows() - r1Count );
   }

   // construct the filtered exposure maps

   if ( booleanParameter("withexpmaps") ) {
	   for ( unsigned ccd = 1; ccd <= 9; ++ccd )
		   for ( unsigned node = 0; node < 2; ++node ) {
			   char extname[9];
			   sprintf( extname, "EXPMAP%u%u", node, ccd );
			   if ( dataset->hasBlock(extname) ) dataset->deleteBlock(extname);
		   }

	   string cmName;
	   if ( booleanParameter("withcombmap") ) {
		   cmName = stringParameter("expimageset");
		   if ( cmName == "" ) cmName = newname(dsName,"EXPMAP","expmap");
	   }

	   addExpMaps( dataset, realParameter("driftbinsize"), cmName, stringParameter("spectrumbinning") );
   }

   /* SOC-SCR-215 */
   std::vector<double> gtiValues;
   for (unsigned int ccdNr = 1 ; ccdNr <=9 ; ccdNr++)
   {
	   std::stringstream s;
       s << "STDGTI" << setw(2) << setfill('0') << ccdNr;
      
       if (dataset->hasBlock(s.str()))
    	   gtiValues.push_back(dataset->table(s.str())->attribute("ONTIME")->asReal());
   }
   double maxOnTime = 0.0;
   for (unsigned int i = 0 ; i < gtiValues.size() ; ++i)
       maxOnTime = max(maxOnTime,gtiValues[i]);

   *(dataset->table("EVENTS"))->addAttribute("ONTIME","Maximum value for all active CCDs","secs") = maxOnTime;
   /* SOC-SCR-215 */

   dataSetServer->close( dataset );
}

void addExpMaps( struct DataSet *dataset,
		 float driftbinsize, const string &combmap, const string spectrumbinning )
{
   const Table *etab	= dataset->table("EVENTS");
   struct GridSpan Enerbinning;

   if (spectrumbinning == "beta")
	   Enerbinning = gridspan( etab->column("BETA_CHANNEL") );
   else
	   Enerbinning = gridspan( etab->column("MLAMBDA_CHANNEL") );

   struct GridSpan xdsp = gridspan( etab->column("XDSP_CHANNEL") );
   float aspfact1	= etab->attribute("ASPFACT1")->asReal();
   int   rejflags	= etab->attribute("REJFLAGS")->asInt();
   double offaxis = etab->attribute("OFFAXIS")->asReal();

   string instrument	= dataset->attribute("INSTRUME")->asString();
   string submode	= dataset->attribute("SUBMODE")->asString();
   int ccdocb		= etab->attribute("CCDOCB")->asInt();

#if 1
   // SUBMODE was not set properly before SAS 5.3; substitute DATAMODE
   if ( submode == "Unknown" )
      submode = dataset->attribute("DATAMODE")->asString();
#endif

   calServer.setInstrument( Xmm::toInstrument(instrument) );
   calServer.state()->mode()->set( Xmm::toMode(submode) );
   calServer.state()->ocb()->set(ccdocb);

   bool saaCorr = false;
   if(etab->hasAttribute("SUNANGCO"))
	   saaCorr = true;

   bool helioCorr = false;
   if (etab->hasAttribute("HELIOFAC"))
	   helioCorr = true;

   struct ExposureMapper mapper( Enerbinning, xdsp, aspfact1,
				 driftbinsize * M_PI / 648000., offaxis, spectrumbinning,
				 saaCorr, helioCorr );

   //Check if the SAA correction has been applied and set the corresponding
   //values
   if(saaCorr)
   {
	   mapper.setSAACorr(etab->attribute("SUNANGLE")->asReal());
	   mapper.setSAACteA(etab->attribute("cteA")->asReal());
	   mapper.setSAASlopeB(etab->attribute("slopeB")->asReal());
   }

   //Check if the heliocentric velocity correction has been applied
   if (etab->hasAttribute("HELIOFAC"))
   {
	   mapper.setHelioCorr(etab->attribute("HELIOFAC")->asReal());
   }

   struct ExposureCombiner *combiner = 0;

   if ( combmap != "" )
      combiner = new ExposureCombiner(combmap,dataset,Enerbinning,xdsp);

   for ( unsigned ccd = 1; ccd <= 9; ++ccd ) {
      calServer.state()->ccd()->set(ccd);
      struct GeometryDataServer *calGDS = 0; calGDS =calServer.getAtom(calGDS);

      char xtabName[9];
      sprintf( xtabName, "EXPOSU0%u", ccd );
      if ( !dataset->hasBlock(xtabName) ) continue;
      struct Table *xtab = dataset->table(xtabName);

      for ( unsigned node = 0; node < 2; ++node ) {
    	  char extname[9], btabName[9], rtabName[9];
    	  sprintf( extname, "EXPMAP%u%u", node, ccd );
    	  sprintf( btabName, "BADPIX%u%u", node, ccd );
    	  sprintf( rtabName, "REJPIX%u%u", node, ccd );

    	  const Table *btab = dataset->table(btabName);
    	  const Table *rtab = dataset->table(rtabName);

    	  mapper.load(xtab,btab,rtab,rejflags,node,calGDS);

    	  longVec size(2);
    	  size[0] = mapper.bounds.row.length();
    	  size[1] = mapper.bounds.col.length();
	 struct Array *array = dataset->addArray( extname, Dal::Real32, size,
						  "s", "exposure(beta,xdsp)" );

#define SET(name,value,label) *array->addAttribute(name,label) = value

	 SET( "CCDNODE", (int)node, "readout node identifier" );
	 SET( "CCDID", (int)ccd, "CCD identifier" );
	 array->addAttribute( etab->attribute("CCDOCB") );
	 array->addAttribute( etab->attribute("SOURCEID") );

	 if ( mapper.groups.size() == 0 ) continue;

	 struct Pixel origin( mapper.bounds.col.min, mapper.bounds.row.min );

	 /* LAMBDA BINNING DEBUG LINES

	 cout << "TEST " << mapper.bounds.col.min
			 << " " << Enerbinning[ mapper.bounds.col.min] + Enerbinning.stride/2.0
			 << " " << mapper.bounds.col.min + 1
			 << " " << Enerbinning[ mapper.bounds.col.min+1]+ Enerbinning.stride/2.0
			 << endl;
	  */
	 string enerLabel;
	 if (stringParameter("spectrumbinning") == "beta")
		 enerLabel = "rad";
	 else
		 enerLabel = "angstrom";

	 SET( "CTYPE1", "", "[dispersion] channel axis type" );
	 SET( "CUNIT1", enerLabel, "channel axis units" );
	 SET( "CRPIX1", 1, "reference channel coordinate" );
	 SET( "CRVAL1", Enerbinning[origin.col], "reference physical coordinate" );
	 SET( "CDELT1", Enerbinning.stride, "uniform channel width" );

	 SET( "CTYPE2", "", "[cross-dispersion] channel axis type" );
	 SET( "CUNIT2", "rad", "channel axis units" );
	 SET( "CRPIX2", 1, "reference channel coordinate" );
	 SET( "CRVAL2", xdsp[origin.row], "reference physical coordinate" );
	 SET( "CDELT2", xdsp.stride, "uniform channel width" );

	 SET( "T_BETA_X", mapper.tBetaX, "transform: chipX per beta bin" );
	 SET( "T_XDSP_X", mapper.tXdspX, "transform: chipX per xdsp bin" );
	 SET( "T_ORIG_X", mapper.tOrigX, "transform: chipX at origin" );
	 SET( "T_BETA_Y", mapper.tBetaY, "transform: chipY per beta bin" );
	 SET( "T_XDSP_Y", mapper.tXdspY, "transform: chipY per xdsp bin" );
	 SET( "T_ORIG_Y", mapper.tOrigY, "transform: chipY at origin" );

	 struct ArrayData *arrayData = array->data();
	 float *buffer = arrayData->real32Data();
	 size_t bufferSize = size[0] * size[1];



	 struct PixelMap<float> xmap( mapper.bounds, 0, buffer );
	 float threshold = 1.788e-7 * mapper.expose(xmap); // 3*2^-24*exposure

	 float datamin, datamax;
	 for ( datamin = datamax = *buffer; bufferSize--; ++buffer ) {
	    if ( fabs(*buffer) < threshold ) *buffer = 0;
	    minimize( datamin, *buffer );
	    maximize( datamax, *buffer );
	 }


	 SET( "DATAMIN", datamin, "minimum data value present" );
	 SET( "DATAMAX", datamax, "maximum data value present" );

#undef SET

	 if ( combiner ) combiner->load( xmap, extname, ccd, node );
	 	 delete arrayData;
      }

      calServer.releaseAtom(calGDS);
   }

   delete combiner;
}

string newname( string model, const string &ptype, const string &alternate )
{
   string::size_type end = model.rfind('.');
   if ( end == string::npos ) end = model.size();

   if ( end >= 10 )
      return model.replace( end-10, 6, ptype );
   else
      return alternate + model.substr(end);
}

/* --- ExposureCombiner methods -------------------------------------------- */

ExposureCombiner::
ExposureCombiner( const string &filename, const DataSet *eventset,
		  const GridSpan &beta, const GridSpan &xdsp )
{
   _dataset = dataSetServer->open( filename, Dal::Create );
   
   _dataset->addAttribute( eventset->attribute("TELESCOP") );
   _dataset->addAttribute( eventset->attribute("INSTRUME") );
   _dataset->addAttribute( eventset->attribute("OBS_ID") );
   _dataset->addAttribute( eventset->attribute("EXP_ID") );
   _dataset->addAttribute( eventset->attribute("DATE-OBS") );
   _dataset->addAttribute( eventset->attribute("DATE-END") );

   if ( eventset->hasAttribute("EXPIDSTR") )
      _dataset->addAttribute( eventset->attribute("EXPIDSTR") );

   longVec size(2);
   size[0] = xdsp.length();
   size[1] = beta.length();

   _array = _dataset->addArray( "PRIMARY", Dal::Real32, size,
				"s", "exposure(beta,xdsp)" );

#define SET(name,value,label) *_array->addAttribute(name,label) = value

   SET( "CTYPE1", "", "[dispersion] channel axis type" );
   SET( "CUNIT1", "rad", "channel axis units" );
   SET( "CRPIX1", 1, "reference channel coordinate" );
   SET( "CRVAL1", beta[beta.min], "reference physical coordinate" );
   SET( "CDELT1", beta.stride, "uniform channel width" );

   SET( "CTYPE2", "", "[cross-dispersion] channel axis type" );
   SET( "CUNIT2", "rad", "channel axis units" );
   SET( "CRPIX2", 1, "reference channel coordinate" );
   SET( "CRVAL2", xdsp[xdsp.min], "reference physical coordinate" );
   SET( "CDELT2", xdsp.stride, "uniform channel width" );

#undef SET

   const Table *events = eventset->table("EVENTS");
   _array->addAttribute( events->attribute("CCDOCB") );
   _array->addAttribute( events->attribute("SOURCEID") );

   _arrayData = _array->data();
   float *buffer = _arrayData->real32Data();
   memset( buffer, 0, size[0] * size[1] * sizeof(float) );
   _map.resize( Bounds(beta,xdsp), 0, buffer );
}

void ExposureCombiner::load( const PixelMap<float> &xmap,
			     const char *name, unsigned ccd, unsigned node )
{
   struct Bounds bounds = xmap.bounds() & _map.bounds();

   if ( !bounds ) {
      errstr << name << " does not overlap the common channel space"
	     << Sas::warning("lostExposure");
      return;
   }

   if ( xmap.bounds().col.min < bounds.col.min )
      errstr << name << " extends " << bounds.col.min - xmap.bounds().col.min
	     << " columns left of the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().col.max > bounds.col.max )
      errstr << name << " extends " << xmap.bounds().col.max - bounds.col.max
	     << " columns right of the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().row.min < bounds.row.min )
      errstr << name << " extends " << bounds.row.min - xmap.bounds().row.min
	     << " rows below the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().row.max > bounds.row.max )
      errstr << name << " extends " << xmap.bounds().row.max - bounds.row.max
	     << " rows above the common channel space"
	     << Sas::warning("lostExposure");

   for ( int row = bounds.row.min; row <= bounds.row.max; ++row ) {
      using Compass::East;
      long src = xmap.id( bounds.col.min - 1, row );
      long dst = _map.id( bounds.col.min - 1, row );

      for ( size_t count = bounds.col.length(); count--; )
    	  _map[ _map.incr(dst,East) ] += xmap[ xmap.incr(src,East) ];
   }

   _nodeid.extend( 2 * ccd + node - 2 );
   _column.extend( xmap.bounds().col.min );
   _column.extend( xmap.bounds().col.max );
}

ExposureCombiner::~ExposureCombiner()
{
#define SET(name,value,label) *_array->addAttribute(name,label) = value

   if ( _nodeid.length() ) {
      float slope = -((float)_nodeid.length()) / _column.length();
      float icept = 1 + _nodeid.max - slope * (_column.min - .5);

      SET( "T_BETA_N", slope, "transform: nodes per beta bin" );
      SET( "T_ORIG_N", icept, "transform: node offset at origin" );
   }

   float datamin, datamax, *buffer = &_map[0];
   size_t bufferSize = _map.bounds().col.length() * _map.bounds().row.length();
   for ( datamin = datamax = *buffer; bufferSize--; ++buffer ) {
      minimize( datamin, *buffer );
      maximize( datamax, *buffer );
   }

   SET( "DATAMIN", datamin, "minimum data value present" );
   SET( "DATAMAX", datamax, "maximum data value present" );

   delete _arrayData;
   dataSetServer->close( _dataset );

#undef SET
}
